@@ -88,7 +88,7 @@ pango_language_get_type (void)
 {
   static GType our_type = 0;
 
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static (I_("PangoLanguage"),
 					     (GBoxedCopyFunc)pango_language_copy,
 					     (GBoxedFreeFunc)pango_language_free);
@@ -193,13 +193,14 @@ _pango_get_lc_ctype (void)
 PangoLanguage *
 pango_language_get_default (void)
 {
-  gchar *lang;
-  PangoLanguage *result;
+  static PangoLanguage *result = NULL;
 
-  lang = _pango_get_lc_ctype ();
-
-  result = pango_language_from_string (lang);
-  g_free (lang);
+  if (G_UNLIKELY (!result))
+    {
+      gchar *lang = _pango_get_lc_ctype ();
+      result = pango_language_from_string (lang);
+      g_free (lang);
+    }
 
   return result;
 }
@@ -415,7 +416,7 @@ script_for_lang_compare (gconstpointer key,
 
 /**
  * pango_language_includes_script:
- * @language: a #PangoLanguage
+ * @language: a #PangoLanguage, or %NULL
  * @script: a #PangoScript
  *
  * Determines if @script is one of the scripts used to
@@ -430,7 +431,8 @@ script_for_lang_compare (gconstpointer key,
  * applications in most circumstances.
  *
  * Return value: %TRUE if @script is one of the scripts used
- * to write @language, or if nothing is known about @language,
+ * to write @language or if nothing is known about @language
+ * (including the case that @language is %NULL),
  * %FALSE otherwise.
  *
  * Since: 1.4
@@ -443,12 +445,10 @@ pango_language_includes_script (PangoLanguage *language,
   unsigned int j;
   const char *lang_str;
 
-  g_return_val_if_fail (language != NULL, FALSE);
-
 #define REAL_SCRIPT(script) \
   ((script) > PANGO_SCRIPT_INHERITED)
 
-  if (!REAL_SCRIPT (script))
+  if (language == NULL || !REAL_SCRIPT (script))
     return TRUE;
 
   lang_str = pango_language_to_string (language);
