@@ -32,30 +32,32 @@
  * @analysis:  #PangoAnalysis structure from pango_itemize()
  * @glyphs:    glyph string in which to store results
  *
- * Given a segment of text and the corresponding 
+ * Given a segment of text and the corresponding
  * #PangoAnalysis structure returned from pango_itemize(),
  * convert the characters into glyphs. You may also pass
  * in only a substring of the item from pango_itemize().
  */
 void
-pango_shape (const gchar      *text, 
-             gint              length, 
-             const PangoAnalysis *analysis,
-             PangoGlyphString *glyphs)
+pango_shape (const gchar      *text,
+	     gint              length,
+	     const PangoAnalysis *analysis,
+	     PangoGlyphString *glyphs)
 {
   int i;
   int last_cluster;
 
+  glyphs->num_glyphs = 0;
+
   if (G_LIKELY (PANGO_IS_ENGINE_SHAPE (analysis->shape_engine) && PANGO_IS_FONT (analysis->font)))
     {
       _pango_engine_shape_shape (analysis->shape_engine, analysis->font,
 				 text, length, analysis, glyphs);
 
       if (G_UNLIKELY (glyphs->num_glyphs == 0))
-        {
+	{
 	  /* If a font has been correctly chosen, but no glyphs are output,
 	   * there's probably something wrong with the shaper.  Trying to be
-	   * informative, we print out the font description, but to not 
+	   * informative, we print out the font description, but to not
 	   * flood the terminal with zillions of the message, we set a flag
 	   * on the font to only err once per font.
 	   */
@@ -63,7 +65,7 @@ pango_shape (const gchar      *text,
 
 	  if (!warned_quark)
 	    warned_quark = g_quark_from_static_string ("pango-shaper-warned");
-	  
+
 	  if (!g_object_get_qdata (G_OBJECT (analysis->font), warned_quark))
 	    {
 	      PangoFontDescription *desc;
@@ -87,13 +89,13 @@ pango_shape (const gchar      *text,
     {
       if (!PANGO_IS_ENGINE_SHAPE (analysis->shape_engine) &&
 	  !_pango_warning_history.shape_shape_engine)
-        {
+	{
 	  _pango_warning_history.shape_shape_engine = TRUE;
 	  g_warning ("pango_shape called with bad shape_engine, expect ugly output");
 	}
       if (!PANGO_IS_FONT (analysis->font) &&
 	  !_pango_warning_history.shape_font)
-        {
+	{
 	  _pango_warning_history.shape_font = TRUE;
 	  g_warning ("pango_shape called with bad font, expect ugly output");
 	}
